class AssistantIdNotSet(Exception):
    """Exception raised when the assistant.id is not set."""

    pass


class ThreadIdNotSet(Exception):
    """Exception raised when the thread.id is not set."""

    pass


class MessageIdNotSet(Exception):
    """Exception raised when the message.id is not set."""

    pass


class RunIdNotSet(Exception):
    """Exception raised when the run.id is not set."""

    pass


class FileIdNotSet(Exception):
    """Exception raised when the file.id is not set."""

    pass
